from .constants import *
from .read import Read
from .genes import Gene
from .intervals import Interval, IntervalsIndex
from .molitem import Molitem
from .counter import ExInCounter
from .transcript import Transcript
from .metadata import MetadataCollection, Metadata
from .neighbors import BalancedKNN, convolve_by_sparse_weights
from .estimation import fit_slope, _fit1_slope, clusters_stats
from .analysis import VelocytoLoom, scatter_viz, ixs_thatsort_a2b



import click
import logging
import sys
from .extract_repeats import extract_repeats
from .extract_intervals import extract_intervals
from .run import run
from .run10x import run10x
from .multi10x import multi10x


@click.group(context_settings=dict(max_content_width=300, terminal_width=300))
def cli() -> None:
    logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)
    return

cli.add_command(extract_repeats)
cli.add_command(extract_intervals)
cli.add_command(run)
cli.add_command(run10x)
cli.add_command(multi10x)



from numpy import zeros
import array
import logging
from typing import *
import velocyto as vcy


class Gene:

    __slots__ = ["genename", "geneid", "chrom", "strand", "lastexon_end_pos", "spliced_mol_counts",
                 "ambiguous_mol_counts", "unspliced_mol_counts", "deduced_tr_end", "start", "end",
                 "read_start_counts_from_locus_end", "ivls", "ivlinside_read_counts", "ivljunction5_read_counts", "ivljunction3_read_counts"]
                 
    def __init__(self, genename: str, geneid: str, chrom: str, strand: str, nbcs: int, genestart: int=0, geneend: int=0) -> None:
        self.genename = genename
        self.geneid = geneid
        self.chrom = chrom
        self.strand = strand
        self.lastexon_end_pos = 0  # Relative to annotated (+ means after)
        self.spliced_mol_counts = zeros(shape=nbcs, dtype=vcy.LOOM_NUMERIC_DTYPE)  # NOTE: this is uint16
        self.ambiguous_mol_counts = zeros(shape=nbcs, dtype=vcy.LOOM_NUMERIC_DTYPE)
        self.unspliced_mol_counts = zeros(shape=nbcs, dtype=vcy.LOOM_NUMERIC_DTYPE)
        self.deduced_tr_end = 0
        self.start = genestart
        self.end = geneend
        self.read_start_counts_from_locus_end = array.array('H')

    def __lt__(self, other: Any) -> bool:
        if self.get_chrom() != other.get_chrom():
            return self.get_chrom() < other.get_chrom()
        elif self.get_strand() != other.get_strand():
            return self.get_strand() < other.get_strand()
        elif self.get_start() != other.get_start():
            return self.get_start() < other.get_start()
        else:
            return self.get_end() < other.get_end()

    def get_start(self) -> int:
        return self.start

    def get_end(self) -> int:
        return self.end

    def get_chrom(self) -> str:
        return self.chrom

    def get_strand(self) -> str:
        return self.strand

    def get_spliced_mol_counts(gene, bcidx: int) -> int:
        """ Molecule counts that are exonic in all transcript models [ordered by barcode index]"""
        return gene.spliced_mol_counts[bcidx]

    def get_ambiguous_mol_counts(gene, bcidx: int) -> int:
        """ Molecule counts that are exonic in some but not all transcript models [ordered by barcode index]"""
        return gene.ambiguous_mol_counts[bcidx]

    def get_unspliced_mol_counts(gene, bcidx: int) -> int:
        """ Molecule counts that are intronic by all transcript models [ordered by barcode index]"""
        return gene.unspliced_mol_counts[bcidx]

    def set_range(self, start: int, end: int) -> None:
        self.start, self.end = start, end

    def set_ivls(self, ivls: List) -> None:
        self.ivls = ivls
        self.ivlinside_read_counts = zeros(shape=len(ivls), dtype="uint32")  # array.array('I', [0] * len(ivls))
        self.ivljunction5_read_counts = zeros(shape=len(ivls), dtype="uint32")  # array.array('I', [0] * len(ivls))
        self.ivljunction3_read_counts = zeros(shape=len(ivls), dtype="uint32")  # added recently

    def num_ivls(self) -> int:
        return len(self.ivls)

    def get_lastexon_counts(self) -> Tuple[int, int]:
        """ return the tuple: (read count spanning most 3' intron-exon junction, read count on most 3' exon) """
        result = 0, 0
        lastidx = len(self.ivls) - 1
        i, step, stop = (0, 1, lastidx) if self.strand == '-' else (lastidx, -1, 0)
        if self.ivls[i].is_sure_exon:
            lastexon_count = self.ivlinside_read_counts[i]
            while i != stop and self.ivls[i + step].is_sure_exon:
                lastexon_count += self.ivljunction5_read_counts[i]
                i += step
                lastexon_count += self.ivlinside_read_counts[i]
            if i != stop and self.ivls[i + step].is_sure_intron:
                lastjunction_count = self.ivljunction5_read_counts[i]
                result = lastjunction_count, lastexon_count
        return result

    def get_lastexon_length(self) -> int:
        i = len(self.ivls) - 1 if self.strand == '+' else 0
        return self.ivls[i].end - self.ivls[i].start

    def get_last_3prime_exon_interval(self) -> int:
        i = len(self.ivls) - 1 if self.strand == '+' else 0
        return self.ivls[i]

    def validate_intron_ivls(self) -> None:
        """ Annotate the introns that can safely be used for intron molecule counting =>
        either flanking exon is sure an exon and has >= 1 read spanning junction """
        for i in range(1, len(self.ivls) - 2):
            if self.ivls[i].is_sure_intron:
                self.ivls[i].is_sure_valid_intron = \
                    self.ivls[i - 1].is_sure_exon and self.ivljunction5_read_counts[i] > 0 or \
                    self.ivls[i + 1].is_sure_exon and self.ivljunction5_read_counts[i + 1] > 0

    def add_read_stats(self, read: vcy.Read) -> None:
        if self.strand == '+':
            self.deduced_tr_end = max(self.deduced_tr_end, read.end() + read.clip3)
            dist_from_end = self.end - read.start()
        else:
            dist_from_end = read.end() - self.start
            if self.deduced_tr_end == 0:
                self.deduced_tr_end = read.start() - read.clip5
            else:
                self.deduced_tr_end = min(self.deduced_tr_end, read.start() - read.clip5)
        if dist_from_end >= len(self.read_start_counts_from_locus_end):
            self.read_start_counts_from_locus_end.extend([0] * (dist_from_end + 100))
        c = self.read_start_counts_from_locus_end[dist_from_end]
        if c < vcy.MAX_USHORT:
            self.read_start_counts_from_locus_end[dist_from_end] = c + 1

    def get_deduced_tr_end(self) -> int:
        return self.deduced_tr_end

    def get_tr_end(self) -> int:
        return self.start if self.strand == '-' else self.end



import cython
import numpy as np
cimport numpy as np
from libc.math cimport sqrt, fabs, log10
from cython.parallel import threadid, parallel, prange

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _colDeltaCor(double[:, ::1] e,
                 double[:, ::1] d,
                 double[:, ::1] rm,
                 int num_threads):
    cdef:
        Py_ssize_t rows = e.shape[0]
        Py_ssize_t cols = e.shape[1]
        Py_ssize_t i, j, c, t
        double[:, :, ::1] t_A = np.zeros((num_threads, e.shape[0], e.shape[1])) # np.tile(e, (3,1,1))
        double[:, ::1] t_b = np.array(d.T, order="C")
        double[:, ::1] t_out = np.zeros((num_threads, e.shape[1]))
        double[:, ::1] t_muA = np.zeros((num_threads, e.shape[1]))
        double[:, :, ::1] t_A_mA = np.zeros((num_threads, e.shape[0], e.shape[1]))
        double[:, ::1] t_b_mb = np.zeros((num_threads, d.shape[0]))
        double[:, ::1] t_ssA = np.zeros((num_threads, e.shape[1]))
        double[::1] t_mub = np.zeros(num_threads)
        double[::1] t_ssb = np.zeros(num_threads)
        double[::1] t_tmp = np.zeros(num_threads)
        int thread_id
    
    with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
        for c in prange(cols, schedule='static', num_threads=num_threads):
            t = threadid() # or 
            
            # subtract the cth column
            for j in range(rows):
                for i in range(cols):
                    t_A[t, j, i] = e[j, i] - e[j, c]
            
            #muA = A.mean(0)
            for j in range(rows):
                for i in range(cols):
                    t_muA[t, i] += t_A[t, j, i]
            for i in range(cols):
                t_muA[t, i] = t_muA[t, i] / rows

            # A_mA = A - muA
            for j in range(rows):
                for i in range(cols):
                    t_A_mA[t, j, i] = t_A[t, j, i] - t_muA[t, i]

            # mub = b.mean()
            for j in range(rows):
                t_mub[t] += t_b[c, j]
            t_mub[t] = t_mub[t] / rows

            # b_mb = b - mub
            for j in range(rows):
                t_b_mb[t, j] = t_b[c, j] - t_mub[t]

            # ssA = (A_mA**2).sum(0)
            for j in range(rows):
                for i in range(cols):
                    t_ssA[t, i] += t_A_mA[t, j, i] * t_A_mA[t, j, i]
            for i in range(cols):
                t_ssA[t, i] = 1. / sqrt(t_ssA[t, i])

            # ssb = (b_mb**2).sum()
            for j in range(rows):
                t_ssb[t] += t_b_mb[t, j] * t_b_mb[t, j] # **2
            t_ssb[t] = 1. / sqrt(t_ssb[t])

            # np.dot(b_mb, A_mA)/(np.sqrt(ssA) * np.sqrt(ssb))
            for j in range(rows):
                t_tmp[t] = t_b_mb[t, j] * t_ssb[t]
                for i in range(cols):
                    rm[c, i] += (t_A_mA[t, j, i] * t_ssA[t, i]) * t_tmp[t]
                    
            # Cleanup
            for i in range(cols):
                t_muA[t, i] = 0
            
            t_mub[t] = 0
                
            for i in range(cols):
                t_ssA[t, i] = 0
                
            t_ssb[t] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _colDeltaCorLog10(double[:, ::1] e,
                      double[:, ::1] d,
                      double[:, ::1] rm,
                      int num_threads,
                      double psc):
    cdef:
        Py_ssize_t rows = e.shape[0]
        Py_ssize_t cols = e.shape[1]
        Py_ssize_t i, j, c, t
        double[:, :, ::1] t_A = np.zeros((num_threads, e.shape[0], e.shape[1])) # np.tile(e, (3,1,1))
        double[:, ::1] t_b = np.array(d.T, order="C")
        double[:, ::1] t_out = np.zeros((num_threads, e.shape[1]))
        double[:, ::1] t_muA = np.zeros((num_threads, e.shape[1]))
        double[:, :, ::1] t_A_mA = np.zeros((num_threads, e.shape[0], e.shape[1]))
        double[:, ::1] t_b_mb = np.zeros((num_threads, d.shape[0]))
        double[:, ::1] t_ssA = np.zeros((num_threads, e.shape[1]))
        double[::1] t_mub = np.zeros(num_threads)
        double[::1] t_ssb = np.zeros(num_threads)
        double[::1] t_tmp = np.zeros(num_threads)
        int thread_id
    
    with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
        for c in prange(cols, schedule='static', num_threads=num_threads):
            t = threadid() # or 
            
            # subtract the cth column
            for j in range(rows):
                for i in range(cols):
                    t_tmp[t] = e[j, i] - e[j, c]
                    if t_tmp[t] > 0:
                        t_A[t, j, i] = log10(fabs(t_tmp[t]) + psc)
                    else:
                        t_A[t, j, i] = -log10(fabs(t_tmp[t]) + psc)
                    
            
            #muA = A.mean(0)
            for j in range(rows):
                for i in range(cols):
                    t_muA[t, i] += t_A[t, j, i]
            for i in range(cols):
                t_muA[t, i] = t_muA[t, i] / rows

            # A_mA = A - muA
            for j in range(rows):
                for i in range(cols):
                    t_A_mA[t, j, i] = t_A[t, j, i] - t_muA[t, i]

            # mub = b.mean()
            for j in range(rows):
                t_mub[t] += t_b[c, j]
            t_mub[t] = t_mub[t] / rows

            # b_mb = b - mub
            for j in range(rows):
                t_b_mb[t, j] = t_b[c, j] - t_mub[t]

            # ssA = (A_mA**2).sum(0)
            for j in range(rows):
                for i in range(cols):
                    t_ssA[t, i] += t_A_mA[t, j, i] * t_A_mA[t, j, i]
            for i in range(cols):
                t_ssA[t, i] = 1. / sqrt(t_ssA[t, i])

            # ssb = (b_mb**2).sum()
            for j in range(rows):
                t_ssb[t] += t_b_mb[t, j] * t_b_mb[t, j] # **2
            t_ssb[t] = 1. / sqrt(t_ssb[t])

            # np.dot(b_mb, A_mA)/(np.sqrt(ssA) * np.sqrt(ssb))
            for j in range(rows):
                t_tmp[t] = t_b_mb[t, j] * t_ssb[t]
                for i in range(cols):
                    rm[c, i] += (t_A_mA[t, j, i] * t_ssA[t, i]) * t_tmp[t]
                    
            # Cleanup
            for i in range(cols):
                t_muA[t, i] = 0
            
            t_mub[t] = 0
                
            for i in range(cols):
                t_ssA[t, i] = 0
                
            t_ssb[t] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _colDeltaCorSqrt(double[:, ::1] e,
                      double[:, ::1] d,
                      double[:, ::1] rm,
                      int num_threads,
                      double psc):
    cdef:
        Py_ssize_t rows = e.shape[0]
        Py_ssize_t cols = e.shape[1]
        Py_ssize_t i, j, c, t
        double[:, :, ::1] t_A = np.zeros((num_threads, e.shape[0], e.shape[1])) # np.tile(e, (3,1,1))
        double[:, ::1] t_b = np.array(d.T, order="C")
        double[:, ::1] t_out = np.zeros((num_threads, e.shape[1]))
        double[:, ::1] t_muA = np.zeros((num_threads, e.shape[1]))
        double[:, :, ::1] t_A_mA = np.zeros((num_threads, e.shape[0], e.shape[1]))
        double[:, ::1] t_b_mb = np.zeros((num_threads, d.shape[0]))
        double[:, ::1] t_ssA = np.zeros((num_threads, e.shape[1]))
        double[::1] t_mub = np.zeros(num_threads)
        double[::1] t_ssb = np.zeros(num_threads)
        double[::1] t_tmp = np.zeros(num_threads)
        int thread_id
    
    with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
        for c in prange(cols, schedule='static', num_threads=num_threads):
            t = threadid() # or 
            
            # subtract the cth column
            for j in range(rows):
                for i in range(cols):
                    t_tmp[t] = e[j, i] - e[j, c]
                    if t_tmp[t] > 0:
                        t_A[t, j, i] = log10(fabs(t_tmp[t]) + 1)
                    else:
                        t_A[t, j, i] = -log10(fabs(t_tmp[t]) + 1)
                    
            
            #muA = A.mean(0)
            for j in range(rows):
                for i in range(cols):
                    t_muA[t, i] += t_A[t, j, i]
            for i in range(cols):
                t_muA[t, i] = t_muA[t, i] / rows

            # A_mA = A - muA
            for j in range(rows):
                for i in range(cols):
                    t_A_mA[t, j, i] = t_A[t, j, i] - t_muA[t, i]

            # mub = b.mean()
            for j in range(rows):
                t_mub[t] += t_b[c, j]
            t_mub[t] = t_mub[t] / rows

            # b_mb = b - mub
            for j in range(rows):
                t_b_mb[t, j] = t_b[c, j] - t_mub[t]

            # ssA = (A_mA**2).sum(0)
            for j in range(rows):
                for i in range(cols):
                    t_ssA[t, i] += t_A_mA[t, j, i] * t_A_mA[t, j, i]
            for i in range(cols):
                t_ssA[t, i] = 1. / sqrt(t_ssA[t, i])

            # ssb = (b_mb**2).sum()
            for j in range(rows):
                t_ssb[t] += t_b_mb[t, j] * t_b_mb[t, j] # **2
            t_ssb[t] = 1. / sqrt(t_ssb[t])

            # np.dot(b_mb, A_mA)/(np.sqrt(ssA) * np.sqrt(ssb))
            for j in range(rows):
                t_tmp[t] = t_b_mb[t, j] * t_ssb[t]
                for i in range(cols):
                    rm[c, i] += (t_A_mA[t, j, i] * t_ssA[t, i]) * t_tmp[t]
                    
            # Cleanup
            for i in range(cols):
                t_muA[t, i] = 0
            
            t_mub[t] = 0
                
            for i in range(cols):
                t_ssA[t, i] = 0
                
            t_ssb[t] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _colDeltaCorpartial(double[:, ::1] e,
                        double[:, ::1] d,
                        double[:, ::1] rm,
                        Py_ssize_t[:, ::1] ixs,
                        int num_threads):
    cdef:
        Py_ssize_t rows = e.shape[0]
        Py_ssize_t cols = e.shape[1]
        Py_ssize_t nrndm = ixs.shape[1]
        Py_ssize_t i, j, c, t, n
        double[:, :, ::1] t_A = np.zeros((num_threads, e.shape[0], nrndm)) # np.tile(e, (3,1,1))
        double[:, ::1] t_b = np.array(d.T, order="C")
        double[:, ::1] t_out = np.zeros((num_threads, nrndm))
        double[:, ::1] t_muA = np.zeros((num_threads, nrndm))
        double[:, :, ::1] t_A_mA = np.zeros((num_threads, e.shape[0], nrndm))
        double[:, ::1] t_b_mb = np.zeros((num_threads, d.shape[0]))
        double[:, ::1] t_ssA = np.zeros((num_threads, nrndm))
        double[::1] t_mub = np.zeros(num_threads)
        double[::1] t_ssb = np.zeros(num_threads)
        double[::1] t_tmp = np.zeros(num_threads)
        int thread_id
    
    with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
        for c in prange(cols, schedule='static', num_threads=num_threads):
            t = threadid() # or
            
            # subtract the cth column
            for j in range(rows):
                for n in range(nrndm):
                    i = ixs[c, n]
                    t_A[t, j, n] = e[j, i] - e[j, c]
                    
            
            #muA = A.mean(0)
            for j in range(rows):
                for n in range(nrndm):
                    i = ixs[c, n]
                    t_muA[t, n] += t_A[t, j, n]
            for n in range(nrndm):
                t_muA[t, n] = t_muA[t, n] / rows

            # A_mA = A - muA
            for j in range(rows):
                for n in range(nrndm):
                    t_A_mA[t, j, n] = t_A[t, j, n] - t_muA[t, n]

            # mub = b.mean()
            for j in range(rows):
                t_mub[t] += t_b[c, j]
            t_mub[t] = t_mub[t] / rows

            # b_mb = b - mub
            for j in range(rows):
                t_b_mb[t, j] = t_b[c, j] - t_mub[t]

            # ssA = (A_mA**2).sum(0)
            for j in range(rows):
                for n in range(nrndm):
                    t_ssA[t, n] += t_A_mA[t, j, n] * t_A_mA[t, j, n]
            for n in range(nrndm):
                t_ssA[t, n] = 1. / sqrt(t_ssA[t, n])

            # ssb = (b_mb**2).sum()
            for j in range(rows):
                t_ssb[t] += t_b_mb[t, j] * t_b_mb[t, j] # **2
            t_ssb[t] = 1. / sqrt(t_ssb[t])

            # np.dot(b_mb, A_mA)/(np.sqrt(ssA) * np.sqrt(ssb))
            for j in range(rows):
                t_tmp[t] = t_b_mb[t, j] * t_ssb[t]
                for n in range(nrndm):
                    i = ixs[c, n]
                    rm[c, i] += (t_A_mA[t, j, n] * t_ssA[t, n]) * t_tmp[t]
                    
            # Cleanup
            for n in range(nrndm):
                t_muA[t, n] = 0
            
            t_mub[t] = 0
                
            for n in range(nrndm):
                t_ssA[t, n] = 0
                
            t_ssb[t] = 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _colDeltaCorLog10partial(double[:, ::1] e,
                             double[:, ::1] d,
                             double[:, ::1] rm,
                             Py_ssize_t[:, ::1] ixs,
                             int num_threads,
                             double psc):
    cdef:
        Py_ssize_t rows = e.shape[0]
        Py_ssize_t cols = e.shape[1]
        Py_ssize_t nrndm = ixs.shape[1]
        Py_ssize_t i, j, c, t, n
        double[:, :, ::1] t_A = np.zeros((num_threads, e.shape[0], nrndm)) # np.tile(e, (3,1,1))
        double[:, ::1] t_b = np.array(d.T, order="C")
        double[:, ::1] t_out = np.zeros((num_threads, nrndm))
        double[:, ::1] t_muA = np.zeros((num_threads, nrndm))
        double[:, :, ::1] t_A_mA = np.zeros((num_threads, e.shape[0], nrndm))
        double[:, ::1] t_b_mb = np.zeros((num_threads, d.shape[0]))
        double[:, ::1] t_ssA = np.zeros((num_threads, nrndm))
        double[::1] t_mub = np.zeros(num_threads)
        double[::1] t_ssb = np.zeros(num_threads)
        double[::1] t_tmp = np.zeros(num_threads)
    
    with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
        for c in prange(cols, schedule='static', num_threads=num_threads):
            t = threadid() # or
            
            # subtract the cth column
            for j in range(rows):
                for n in range(nrndm):
                    i = ixs[c, n]
                    t_tmp[t] = e[j, i] - e[j, c]
                    if t_tmp[t] > 0:
                        t_A[t, j, n] = log10(fabs(t_tmp[t]) + psc)
                    else:
                        t_A[t, j, n] = -log10(fabs(t_tmp[t]) + psc)
                    
            
            #muA = A.mean(0)
            for j in range(rows):
                for n in range(nrndm):
                    i = ixs[c, n]
                    t_muA[t, n] += t_A[t, j, n]
            for n in range(nrndm):
                t_muA[t, n] = t_muA[t, n] / rows

            # A_mA = A - muA
            for j in range(rows):
                for n in range(nrndm):
                    t_A_mA[t, j, n] = t_A[t, j, n] - t_muA[t, n]

            # mub = b.mean()
            for j in range(rows):
                t_mub[t] += t_b[c, j]
            t_mub[t] = t_mub[t] / rows

            # b_mb = b - mub
            for j in range(rows):
                t_b_mb[t, j] = t_b[c, j] - t_mub[t]

            # ssA = (A_mA**2).sum(0)
            for j in range(rows):
                for n in range(nrndm):
                    t_ssA[t, n] += t_A_mA[t, j, n] * t_A_mA[t, j, n]
            for n in range(nrndm):
                t_ssA[t, n] = 1. / sqrt(t_ssA[t, n])

            # ssb = (b_mb**2).sum()
            for j in range(rows):
                t_ssb[t] += t_b_mb[t, j] * t_b_mb[t, j] # **2
            t_ssb[t] = 1. / sqrt(t_ssb[t])

            # np.dot(b_mb, A_mA)/(np.sqrt(ssA) * np.sqrt(ssb))
            for j in range(rows):
                t_tmp[t] = t_b_mb[t, j] * t_ssb[t]
                for n in range(nrndm):
                    i = ixs[c, n]
                    rm[c, i] += (t_A_mA[t, j, n] * t_ssA[t, n]) * t_tmp[t]
                    
            # Cleanup
            for n in range(nrndm):
                t_muA[t, n] = 0
            
            t_mub[t] = 0
                
            for n in range(nrndm):
                t_ssA[t, n] = 0
                
            t_ssb[t] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _colDeltaCorSqrtpartial(double[:, ::1] e,
                             double[:, ::1] d,
                             double[:, ::1] rm,
                             Py_ssize_t[:, ::1] ixs,
                             int num_threads,
                             double psc):
    cdef:
        Py_ssize_t rows = e.shape[0]
        Py_ssize_t cols = e.shape[1]
        Py_ssize_t nrndm = ixs.shape[1]
        Py_ssize_t i, j, c, t, n
        double[:, :, ::1] t_A = np.zeros((num_threads, e.shape[0], nrndm)) # np.tile(e, (3,1,1))
        double[:, ::1] t_b = np.array(d.T, order="C")
        double[:, ::1] t_out = np.zeros((num_threads, nrndm))
        double[:, ::1] t_muA = np.zeros((num_threads, nrndm))
        double[:, :, ::1] t_A_mA = np.zeros((num_threads, e.shape[0], nrndm))
        double[:, ::1] t_b_mb = np.zeros((num_threads, d.shape[0]))
        double[:, ::1] t_ssA = np.zeros((num_threads, nrndm))
        double[::1] t_mub = np.zeros(num_threads)
        double[::1] t_ssb = np.zeros(num_threads)
        double[::1] t_tmp = np.zeros(num_threads)
    
    with nogil, cython.boundscheck(False), cython.wraparound(False), cython.cdivision(True):
        for c in prange(cols, schedule='static', num_threads=num_threads):
            t = threadid() # or
            
            # subtract the cth column
            for j in range(rows):
                for n in range(nrndm):
                    i = ixs[c, n]
                    t_tmp[t] = e[j, i] - e[j, c]
                    if t_tmp[t] > 0:
                        t_A[t, j, n] = sqrt(fabs(t_tmp[t]) + psc)
                    else:
                        t_A[t, j, n] = -sqrt(fabs(t_tmp[t]) + psc)
                    
            
            #muA = A.mean(0)
            for j in range(rows):
                for n in range(nrndm):
                    i = ixs[c, n]
                    t_muA[t, n] += t_A[t, j, n]
            for n in range(nrndm):
                t_muA[t, n] = t_muA[t, n] / rows

            # A_mA = A - muA
            for j in range(rows):
                for n in range(nrndm):
                    t_A_mA[t, j, n] = t_A[t, j, n] - t_muA[t, n]

            # mub = b.mean()
            for j in range(rows):
                t_mub[t] += t_b[c, j]
            t_mub[t] = t_mub[t] / rows

            # b_mb = b - mub
            for j in range(rows):
                t_b_mb[t, j] = t_b[c, j] - t_mub[t]

            # ssA = (A_mA**2).sum(0)
            for j in range(rows):
                for n in range(nrndm):
                    t_ssA[t, n] += t_A_mA[t, j, n] * t_A_mA[t, j, n]
            for n in range(nrndm):
                t_ssA[t, n] = 1. / sqrt(t_ssA[t, n])

            # ssb = (b_mb**2).sum()
            for j in range(rows):
                t_ssb[t] += t_b_mb[t, j] * t_b_mb[t, j] # **2
            t_ssb[t] = 1. / sqrt(t_ssb[t])

            # np.dot(b_mb, A_mA)/(np.sqrt(ssA) * np.sqrt(ssb))
            for j in range(rows):
                t_tmp[t] = t_b_mb[t, j] * t_ssb[t]
                for n in range(nrndm):
                    i = ixs[c, n]
                    rm[c, i] += (t_A_mA[t, j, n] * t_ssA[t, n]) * t_tmp[t]
                    
            # Cleanup
            for n in range(nrndm):
                t_muA[t, n] = 0
            
            t_mub[t] = 0
                
            for n in range(nrndm):
                t_ssA[t, n] = 0
                
            t_ssb[t] = 0


from typing import *
import velocyto as vcy


class Transcript:
    """Object rappresenting a transcript"""
    __slots__ = ["trname", "trid", "genename", "geneid", "chrom", "strand", "_exons", "exons_sorted"]
    
    def __init__(self, trname: str, trid: str, genename: str, geneid: str, chrom: str, strand: str) -> None:
        """Object rappresenting a transcript
        
        Attributes
        ----------
        trname: str, trid: str, genename: str, geneid: str, chrom: str, strand: str

        Methods
        -------
        get_chromstrand, add_exon, fuse, sorted_exons, sorted_ivls
        get_start, get_end, get_chrom, get_strand, is_fw, get_trname, get_genename, get_trid
        spans_over
        
        Note
        ----
        It only stores the exons, the introns are inferred as exon_n_end->exon_n+1_start
        """
        self.trname = trname
        self.trid = trid
        self.genename = genename
        self.geneid = geneid
        self.chrom = chrom
        self.strand = strand
        self._exons = []  # type: List[List[int]]

    def __lt__(self, other: Any) -> bool:
        if self.get_chrom() != other.get_chrom():
            return self.get_chrom() < other.get_chrom()
        elif self.get_strand() != other.get_strand():
            return self.get_strand() < other.get_strand()
        elif self.get_start() != other.get_start():
            return self.get_start() < other.get_start()
        else:
            return self.get_end() < other.get_end()

    def get_chromstrand(self) -> str:
        return self.chrom + self.strand

    def add_exon(self, start: int, end: int, fuse: bool=False) -> None:
        self._exons.append([start, end])
        self.exons_sorted = False
        if fuse:
            self.fuse()
    
    def fuse(self) -> None:
        self._exons.sort()
        self.exons_sorted = True
        i = 1
        while i < len(self._exons):
            if self._exons[i][0] <= self._exons[i - 1][1]:
                self._exons[i - 1][1] = self._exons[i][1]
                del self._exons[i]
            else:
                i += 1

    def sorted_exons(self) -> List[Any]:
        self._exons.sort()
        self.exons_sorted = True
        return self._exons
    
    def sorted_ivls(self) -> Iterable:
        """Returns sorted intervals of both exons and introns using only the exons
        """
        if len(self._exons) > 0:
            exonstart, exonend = self._exons[0][0:2]
            yield [exonstart, exonend, vcy.EXON]
            for i in range(1, len(self._exons)):
                exonstart = self._exons[i][0]
                yield [exonend, exonstart, vcy.INTRON]
                exonend = self._exons[i][1]
                yield [exonstart, exonend, vcy.EXON]

    def spans_over(self, interval: Tuple[int, int]) -> bool:
        return (self.get_start() < interval[-1]) and (self.get_end() > interval[0])

    def get_start(self) -> int:
        return self.sorted_exons()[0][0]

    def get_end(self) -> int:
        return self.sorted_exons()[-1][1]

    def get_chrom(self) -> str:
        return self.chrom

    def get_strand(self) -> str:
        return self.strand

    def is_fw(self) -> bool:
        return self.strand == '+'

    def get_trname(self) -> str:
        return self.trname

    def get_genename(self) -> str:
        return self.genename

    def get_trid(self) -> str:
        return self.trid

    def get_geneid(self) -> str:
        return self.geneid

    def __str__(self) -> Any:
        return "Transcript(TrID=%s TrName=%s GeneID=%s GeneName=%s Chrom=%s Strand=%s\n\tExons=%s)\ni\tIvls=%s" % \
            (self.trid, self.trname, self.geneid, self.genename, self.chrom, self.strand,
             ",".join("%s-%s" % (s[0], s[1]) for s in self.sorted_exons()),
             ",".join("%s-%s:%s" % (s[0], s[1], s[2]) for s in self.sorted_ivls()))



import logging
import copy
import os
import re
import sys
import click
import gzip
import bisect
import subprocess
from collections import defaultdict
from typing import *
import velocyto as vcy

logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


@click.command(short_help="Transform a repeats .gtf file into a intervals .gtf file.")
@click.argument("gtffile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.option("--tolerance", "-t",
              default=5,
              show_default=True)
def extract_repeats(gtffile: str, tolerance: int) -> None:
    """Transform a repeats .gtf file into a intervals .gtf file.

    GTFFILE  .gtf file with annotated repeats to mask
    """
    
    sorted_filename = gtffile.split(".")[0] + "_sorted.gtf"
    logging.debug(f"Sorting by `sort -k1,1 -k7,7 -k4,4n {gtffile} > {sorted_filename}`")
    with open(sorted_filename, "w") as f:
        p1 = subprocess.run(["sort", "-k1,1", "-k7,7", "-k4,4n", gtffile],
                            stdout=f)
        
    # Parse arguments
    logging.debug(f'Opening {sorted_filename}, output will be at {gtffile.split(".")[0] + "_joined.gtf"}')
    fin = open(sorted_filename)
    fout = open(gtffile.split(".")[0] + "_joined.gtf", "w")

    curr_chromstrand = ""
    curr_end = -1000
    TOLERANCE = tolerance
    headerlines = []
    for line in fin:
        if line.startswith('#'):
            headerlines.append(line)
            continue
        
        fields = line.rstrip().split('\t')
        chrom, feature_class, feature_type, start, end, junk, strand, junk, tags = fields
        start = int(start)
        end = int(end)
        chromstrand = chrom + strand
        
        curr_chrom: str
        curr_feature_class: str
        curr_feature_type: str
        curr_n: int
        curr_start: int
        curr_tags: str
        # If there is no overlap with the next
        if chromstrand != curr_chromstrand or start > curr_end + TOLERANCE:
            if curr_chromstrand != "":
                new_entry = (curr_chrom, curr_feature_class, curr_feature_type,
                             str(curr_start), str(curr_end), ".", strand, ".",
                             f"number_elements {curr_n}; {curr_tags}")
                fout.write("\t".join(new_entry).strip() + "\n")
            
            curr_chrom = chrom
            curr_feature_class = feature_class
            curr_feature_type = feature_type
            curr_start = start
            curr_end = end
            curr_n = 1
            curr_strand: str = strand
            curr_tags = tags
            curr_chromstrand = chromstrand
        else:
            gap = start - curr_end
            curr_end = end
            curr_n += 1
            curr_tags = f"{curr_tags} gap {gap}; {tags}" if gap > 0 else curr_tags + tags
    fout.close()
    fin.close()

    logging.debug(f"Deleting the sorted file {sorted_filename}")
    os.remove(sorted_filename)



from typing import *
from collections import defaultdict
import logging
import velocyto as vcy


class Interval:
    """ Holds an exon/intron interval read from the interval file """
    # Since many of this object are instanciated, implementing slots make it more efficient
    __slots__ = ['start', 'end', 'gene', "ivlidx", "ivltype", "is_sure_exon", "is_maybe_exon", "is_sure_intron", "is_sure_valid_intron", "is_last3prime"]

    def __init__(self, start: str, end: str, gene: vcy.Gene, ivlidx: int, ivltype: str, is_last3prime: bool) -> None:
        self.start, self.end, self.gene, self.ivlidx, self.ivltype = int(start), int(end), gene, ivlidx, ivltype
        self.is_last3prime = is_last3prime
        self.is_maybe_exon = ivltype in ('Eei', 'E-i')
        self.is_sure_exon = ivltype in ('E--', 'Ee-')
        self.is_sure_intron = ivltype in ('I--', 'I-i')
        self.is_sure_valid_intron = self.is_sure_intron  # NOTE This is just initialized and it will be adjusted later based on junction analysis
    
    def add_read_inside(self) -> None:
        self.gene.ivlinside_read_counts[self.ivlidx] += 1

    def add_read_spanning5end(self) -> None:
        self.gene.ivljunction5_read_counts[self.ivlidx] += 1

    def add_read_spanning3end(self) -> None:
        self.gene.ivljunction3_read_counts[self.ivlidx] += 1

    def ends_upstream_of(self, read: vcy.Read) -> bool:
        """The following situation happens
                                                            Read
                                               *|||segment|||-?-||segment|||????????
                ???????|||||Ivl|||||||||*

        """
        return self.end < read.pos

    def starts_upstream_of(self, segment: Tuple[int, int]) -> bool:
        """The following situation happens

                            *||||||segment|||||????????
                *|||||||||||||Ivl||||||||||????????????

        """
        return self.start < segment[-1]

    def contains(self, segment: Tuple[int]) -> bool:
        """The following situation happens

                     *||||||segment|||||*
                *|||||||||||||Ivl||||||||||||||||*

        """
        return (segment[0] >= self.start) and (segment[-1] <= self.end)

    def start_overlaps_with_part_of(self, segment: Tuple[int, int], minimum_flanking: int=vcy.MIN_FLANK) -> bool:
        """The following situation happens

          *---|||segment||---*
                *|||||||||||||Ivl||||||||||||||||*

        where `---` idicates the minimum flanking

        """
        return (segment[0] + minimum_flanking < self.start) and (segment[-1] - minimum_flanking > self.start)

    def end_overlaps_with_part_of(self, segment: Tuple[int, int], minimum_flanking: int=vcy.MIN_FLANK) -> bool:
        """The following situation happens

                                      *---|||segment||---*
                *|||||||||||||Ivl||||||||||||||||*

        where `---` idicates the minimum flanking
            
        """
        return (segment[0] + minimum_flanking < self.end) and (segment[-1] - minimum_flanking > self.end)

    def __lt__(self, other: Any) -> bool:
        assert type(other) == Interval, "Comparison can be only between two intervals"
        if self.start == other.start:
            return self.end < other.end
        return self.start < other.start

    def __gt__(self, other: Any) -> bool:
        assert type(other) == Interval, "Comparison can be only between two intervals"
        if self.start == other.start:
            return self.end > other.end
        return self.start > other.start


class IntervalsIndex:
    """ Search help class used to find the intervals that a read is spanning """
    def __init__(self, ivls: List[Interval]) -> None:
        self.ivls = ivls
        self.ivls.sort()  # NOTE: maybe I am sorting twice check what I do upon creation
        self.iidx = 0  # index of the current interval
        self.maxiidx = len(ivls) - 1

    @ property
    def last_interval_not_reached(self) -> bool:
        return self.iidx < self.maxiidx

    def has_ivls_enclosing(self, read: vcy.Read) -> bool:
        """Finds out if there are intervals that are fully containing all the read segments

        Args
        ----
        read: vcy.Read
            the read object to be analyzed

        Returns
        -------
        respones: bool
            if one has been found

        """
        if len(self.ivls) == 0:
            return False

        ivl = self.ivls[self.iidx]  # current interval
        
        # Move forward until we find the position we will never search left of again (because the reads are ordered)
        while self.last_interval_not_reached and ivl.ends_upstream_of(read):
            # move to the next interval
            self.iidx += 1
            ivl = self.ivls[self.iidx]
        
        for segment in read.segments:
            segment_matchtype = 0
            # Local search for each segment move a little forward (this just moves a coupple of intervals)
            i = self.iidx
            while i < self.maxiidx and ivl.starts_upstream_of(segment):
                matchtype = 0  # No match
                if ivl.contains(segment):
                    matchtype = vcy.MATCH_INSIDE
                if ivl.start_overlaps_with_part_of(segment):  # NOTE: should this be elif or it makes sense to allow both?
                    matchtype |= vcy.MATCH_OVER5END
                if ivl.end_overlaps_with_part_of(segment):  # NOTE: should this be elif or it makes sense to allow both?
                    matchtype |= vcy.MATCH_OVER3END
                
                segment_matchtype |= matchtype
                # move to the next interval
                i += 1
                ivl = self.ivls[i]
            
            # If one of the segments does not match inside a repeat we return false
            if segment_matchtype ^ vcy.MATCH_INSIDE:
                return False
        # If I arrive at this point of the code all the segment matched inside
        return True
        
    def find_overlapping_ivls(self, read: vcy.Read) -> Tuple[set, Dict[Any, int]]:
        """Finds the overlap between Read and intervals

        Args
        ----
        read: vcy.Read
            the read object to be analyzed

        Returns
        -------
        matchgenes: set
            the genes that the read is overlapping with

        matchivls: dict: {vcy.Interval: int}
            A dictionary with keys the intervals that the read is overlapping with and values the kind of overlapping
            it is one of vcy.MATCH_INSIDE (1), vcy.MATCH_OVER5END (2), vcy.MATCH_OVER3END (4)

        """
        matchgenes = set()  # type: set
        matchivls = defaultdict(int)  # type: Dict[Any, int]
        if len(self.ivls) == 0:
            # logging.warn("IntervalsIndex %s contains no intervals" % self)
            return matchgenes, matchivls

        ivl = self.ivls[self.iidx]  # current interval
        
        # Move forward until we find the position we will never search left of again (because the reads are ordered)
        while self.last_interval_not_reached and ivl.ends_upstream_of(read):
            # move to the next interval
            self.iidx += 1
            ivl = self.ivls[self.iidx]
        
        # Loop trough the mapping segments of a read (e.g. just one of an internal exon, generally 2 for a splice or intron)
        for segment in read.segments:
            
            # Local search for each segment move a little forward (this just moves a coupple of intervals)
            i = self.iidx
            while i < self.maxiidx and ivl.starts_upstream_of(segment):
                matchtype = 0  # No match
                if ivl.contains(segment):
                    matchtype = vcy.MATCH_INSIDE
                if ivl.start_overlaps_with_part_of(segment):  # NOTE: should this be elif or it makes sense to allow both?
                    matchtype |= vcy.MATCH_OVER5END
                if ivl.end_overlaps_with_part_of(segment):  # NOTE: should this be elif or it makes sense to allow both?
                    matchtype |= vcy.MATCH_OVER3END
                if matchtype:
                    # if there is a segment matching add it to the set of genes
                    matchgenes.add(ivl.gene)
                    # and add the match type to the set of genes
                    matchivls[ivl] |= matchtype
                # move to the next interval
                i += 1
                ivl = self.ivls[i]
    
        return matchgenes, matchivls



import logging
import copy
import os
import click
import re
import sys
import gzip
import bisect
from collections import defaultdict
from typing import *
import velocyto as vcy

logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


@click.command(short_help="Transform a genome annotation .gtf file into a intervals .txt file.")
@click.option('--dogenes/--no-dogenes',
              default=True,
              help="whether to process gene models",
              show_default=True)
@click.option('--dotrs/--no-dotrs',
              default=True,
              help="whether to process transcript models",
              show_default=True)
@click.option("--outfileprefix", "-p",
              help="prefix to the output files",
              type=click.Path(resolve_path=True),
              default=os.path.realpath("./intervals"),
              show_default=True)
@click.argument("gtffile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
def extract_intervals(dogenes: bool, dotrs: bool, outfileprefix: str, gtffile: str) -> None:
    """Transform a genome annotation .gtf file into a intervals .txt file required to run velocyto.

    GTF_FILE: input file

    """
    dolog = False

    # Parse arguments
    infile = gtffile
    if outfileprefix is None:
        outfileprefix = os.path.splitext(infile)[0]
    
    # Initialize the containers
    genes = {}  # type: Dict[str, vcy.Gene]
    trs = {}  # type: Dict[str, vcy.Transcript]
    trs_by_chromstrand = defaultdict(list)  # type: Dict[str, List]

    def fix_othertrset(tr: vcy.Transcript, otrset: List) -> Tuple[int, List]:
        if len(otrset) == 0:
            return 0, []
        otrnames = []
        othergene = 1  # NOTE this is mostly for transcript model analysis (should almost never happen if gene model analysis is used)
        for otr in otrset:
            if otr.get_genename() != tr.get_genename():  # NOTE IMPORTANT we should choose if it is get_genename or get_geneid
                othergene = 2
            otrnames.append(otr.get_trname())
        otrnames.sort()
        return othergene, otrnames

    headerlines: List[Any] = []

    # Precompile regex for speed
    regex_trid = re.compile('transcript_id "([^"]+)"')
    regex_trname = re.compile('transcript_name "([^"]+)"')
    regex_geneid = re.compile('gene_id "([^"]+)"')
    regex_genename = re.compile('gene_name "([^"]+)"')
    regex_exonno = re.compile('exon_number "([^"]+)"')

    # Loop trough the .gtf file and get transcript models and assemble gene models by fusing exons
    # NOTE: the parsing could be done using pysam tabix
    logging.debug(f"Reading {infile}")
    for nth_line, line in enumerate(open(infile)):
        if line.startswith('#'):
            headerlines.append(line)
            continue
        
        fields = line.rstrip().split('\t')
        chrom, feature_class, feature_type, start, end, junk, strand, junk, tags = fields
        # Deal with possible incongruences between .gtf and bam file chromosome naming
        if "chr" in chrom[:4]:
            chrom = chrom[3:].split(".")[0]
        else:
            chrom = chrom.split(".")[0]
        if feature_type in ("exon"):
            trid = regex_trid.search(tags).group(1)
            trname = regex_trname.search(tags).group(1)
            geneid = regex_geneid.search(tags).group(1)
            genename = regex_genename.search(tags).group(1)
            exonno = regex_exonno.search(tags).group(1)
            start = int(start)
            end = int(end)
            chromstrand = chrom + strand
            # If encountering this transcript for the first time, create the object and add it to containers
            if trname not in trs:
                tr_tmp = vcy.Transcript(trname, trid, genename, geneid, chrom, strand)  # type: vcy.Transcript
                trs[trname] = tr_tmp
                trs_by_chromstrand[chromstrand].append(tr_tmp)
            # Add the exon to the transcript
            trs[trname].add_exon(start, end)
            # If encountering the gene of this transcript for the first time, create the object and add it to containers
            if geneid not in genes:
                gene = vcy.Transcript(genename, geneid, genename, geneid, chrom, strand)
                genes[geneid] = gene
            genes[geneid].add_exon(start, end, fuse=True)  # In this case fuse the exons if overlapping

        if nth_line % 100000 == 0:
            logging.debug(f"Read {nth_line} lines of {infile}")

    logging.debug(f"Completed reading data for {len(trs)} transcripts and {len(genes)} genes")

    #  Make containers of Transcript by chromosome+strand and bins of 100KB
    binnedtrs_by_chromstrand = {}  # type: Dict[str, Dict[int, set]]
    binsize: int = vcy.BINSIZE_BP  # The bin is used to look up for other transcripts
    for chromstrand, trlist in trs_by_chromstrand.items():
        binnedtrs = defaultdict(set)  # type: Dict[int, set]
        for t in trlist:
            for trbin in range(t.get_start() // binsize, 1 + t.get_end() // binsize):
                binnedtrs[trbin].add(t)
        binnedtrs_by_chromstrand[chromstrand] = binnedtrs

    features = []  # type: List
    if dogenes:
        features.append(('gene', genes))
    if dotrs:
        features.append(('tr', trs))
    for name, modeldict in features:
        outfile = f"{outfileprefix}_{name}_ivls.txt"
        fd = open(outfile, 'w')
        fd.write(f"# Generated from {infile}\n")
        fd.write("".join(headerlines))
        fd.write("# E: interval is exon in the models, I: interval is intron in the model\n")
        fd.write("# last two fields show other transcripts that are exon and introns within the same interval\n")
        fd.write("#TrName\tTrID\tGeneName\tGeneID\tChrom\tStrand\tIntervals\n")
        models = list(modeldict.values())  # type: List[vcy.Transcript]
        models.sort()

        logging.debug("Processing %s %ss..." % (len(models), name))
        for n_tr, tr in enumerate(models):
            chromstrand = tr.get_chromstrand()
            trstart, trend = tr.get_start(), tr.get_end()
            trname, trid, trgenename, trgeneid = tr.get_trname(), tr.get_trid(), tr.get_genename(), tr.get_geneid()

            # Retrieve transcripts within the same bin
            othertrs = set()
            for trbin in range(trstart // binsize, 1 + trend // binsize):
                othertrs.update(binnedtrs_by_chromstrand[chromstrand][trbin])
            if dolog and tr.get_trname() == "Lama4-003":
                logging.debug(tr)

            # Not used right now
            # Extend the 5' or 3' end respecting the presence of other transcripts
            for now5end, extlen in [(True, vcy.EXTENSION5_LEN), (False, vcy.EXTENSION3_LEN)]:
                if extlen > 0:
                    extension_ok = True
                    extstart, extend = (trstart - extlen, trstart) if tr.is_fw() == now5end else (trend, trend + extlen)
                    for otr in othertrs:
                        if otr.get_geneid() != trgeneid and otr.get_start() < extend and otr.get_end() > extstart:
                            for oivl in otr.sorted_exons():
                                if oivl[0] < extend and oivl[1] > extstart:
                                    logging.debug(f"Extension of {trname} hindered by {otr.get_trname()}")
                                    extension_ok = False
                        if not extension_ok:
                            break
                    if extension_ok:
                        if tr.is_fw() == now5end:
                            trstart = extstart
                        else:
                            trend = extend

            # Mark up every base in transcript locus that is own or other transcript's exon/intron
            
            # Initialize a list of list (with a list for every base)
            trlen = trend - trstart
            profile = []  # type: List[List[Any]]
            # every entry of profile constains [type_own_interval, set_other_transcripts_exon_overlap, set_other_transcripts
            for i in range(trlen):
                profile.append([0, set(), set()])  # NOTE: maybe we can implement this with numpy arrays dtype=object
            
            # Loop through the intervals (intron and exons) and mark every base with its interval type
            for ivl in tr.sorted_ivls():
                ivlstart, ivlend, ivltype = ivl  # type: Tuple[int, int, int]
                if dolog and trname == "Lama4-003":  # NOTE: I might want to get rid of this dolog checks completelly
                    logging.debug("Marking", ivlstart, ivlend, ivltype)
                for i in range(ivlstart - trstart, min(trlen, ivlend - trstart)):
                    profile[i][0] = ivltype

            # check other transctipts in the window/bin
            for otr in othertrs:
                if otr == tr:
                    continue
                # if there is some possible overlap between the two transcripts
                if otr.spans_over((trstart, trend)):
                    if dolog and trname == "Lama4-003":
                        logging.debug("Overloading %s" % otr.get_trname())
                    
                    # Visit in order all the intervals and mark the overlappings
                    for oivl in otr.sorted_ivls():
                        ostart, oend, otype = oivl
                        if ostart > trend:
                            break  # this interval is downsream the tr: stop looping
                        if oend < trstart:
                            continue  # this interval is upstream the tr: not interesting
                        pidx = 1 if otype == vcy.EXON else 2
                        if dolog and trname == "Lama4-003":
                            logging.debug("              ivl:", otr.get_trname(), ostart, oend, pidx)
                        # Add the other transcript to the profile
                        for i in range(max(0, ostart - trstart), min(trlen, oend - trstart)):
                            profile[i][pidx].add(otr)
        
            # Now delineate intervals on transcript locus that are certainly own exon, certainly own intron, maybe own exon or other
            currp = [None, None, None]  # type: List[Any]
            outivls = []  # type: List[List[Any]]
            pos = 0
            currivlstart: int
            otherexons: Any
            otherintrons: Any
            ivltype_str: str
            for p in profile:
                if p != currp:
                    if currp[0] is not None:  # first time skip
                        outivls.append([currivlstart, pos + trstart, ivltype_str, otherexons, otherintrons])
                    currp = p
                    ivltype_str = 'E' if currp[0] == vcy.EXON else 'I'
                    otherexonstatus, otherexons = fix_othertrset(tr, p[1])
                    ivltype_str += '-eE'[otherexonstatus]
                    otherintronstatus, otherintrons = fix_othertrset(tr, p[2])
                    ivltype_str += '-iI'[otherintronstatus]
                    currivlstart = pos + trstart
                pos += 1
            if currp[0] is not None:  # do it a last time for the last iteration
                outivls.append([currivlstart, pos + trstart, ivltype_str, otherexons, otherintrons])
        
            # Write output file
            fd.write("%s\t%s\t%s\t%s\t%s\t%s" % (trname, trid, trgenename, trgeneid, tr.get_chrom(), tr.get_strand()))
            for ivl in outivls:
                fd.write("\t%s-%s:%s:%s:%s" % (ivl[0], ivl[1], ivl[2], ",".join(ivl[3]), ",".join(ivl[4])))
            fd.write("\n")

            if (n_tr + 1) % 2500 == 0:
                logging.debug(f"{n_tr+1} {name}s already processed")

        fd.close()

    outputs_names = ""
    for name, modeldict in features:
        outputs_names += f"{outfileprefix}_{name}_ivls.txt"
    logging.debug(f"Done without errors. Find the outputs at {outputs_names}")



from typing import *
import inspect


class Read:
    """ Container for reads from sam alignment file """
    __slots__ = ["bc", "umi", "chrom", "strand", "pos", "cigar", "segments", "clip5", "clip3", "spliced"]

    def __init__(self, bc: str, umi: str, chrom: str, strand: str, pos: int, cigar: str, segments: List, clip5: Any, clip3: Any, spliced: bool) -> None:
        self.bc, self.umi, self.chrom, self.strand, self.pos, self.cigar, self.segments, self.clip5, self.clip3, self.spliced = \
            bc, umi, chrom, strand, pos, cigar, segments, clip5, clip3, spliced

    def is_spliced(self) -> bool:
        return self.spliced  # len(self.segments) > 1
        
    def start(self) -> int:
        return self.segments[0][0]

    def end(self) -> int:
        return self.segments[-1][1]

    def __str__(self) -> str:
        tmp = ""
        for i in self.__slots__:
            attribute = getattr(self, i)
            tmp += f"{i}: {type(attribute).__name__}={attribute}, "
        return tmp



from typing import *
import velocyto as vcy
import numpy as np


class Molitem:
    """ Keeps track of the alignments (to intervals of one gene) of all reads corresponding to the same molecule """

    __slots__ = ["gene", "has_some_spliced_read", "ivlhits"]

    def __init__(self, gene: vcy.Gene) -> None:
        self.gene = gene
        self.has_some_spliced_read = False  # type: bool
        self.ivlhits = np.zeros(gene.num_ivls(), dtype=bool)

    def mark_hit_ivls(self, matchivls: List[vcy.Interval], read_is_spliced: bool = False) -> None:
        """ Add info for the alignment(s) of one read """
        for matchivl in matchivls:
            self.ivlhits[matchivl.ivlidx] = True
        self.has_some_spliced_read = self.has_some_spliced_read or read_is_spliced

    def count(self, my_bcidx: int) -> None:
        """ Call after all reads have been processed to annotate this molecule on the gene """
        num_sure_exon_ivls = num_maybe_exon_ivls = 0
        # Loop trough the gene's intervals
        for hit, ivl in zip(self.ivlhits, self.gene.ivls):
            if hit:
                # if there is at least one evidence of sure intron mapping count as a unspliced molecule
                if ivl.is_sure_valid_intron:
                    self.gene.unspliced_mol_counts[my_bcidx] += 1
                    break
                num_sure_exon_ivls += ivl.is_sure_exon
                num_maybe_exon_ivls += ivl.is_maybe_exon
                # NOTE here we could keep track of ivl.is_sure_intron for the purpose of excluding them in the following condition
        if (num_sure_exon_ivls > 0) or self.has_some_spliced_read:  # NOTE: this condition might be reconsidered
            # alternative might be if (num_sure_exon_ivls > 0 and num_maybe_exon_ivls == 0) or self.has_some_spliced_read:  #
            # Either all reads hit to sure exons, or all reads hit to sure/maybe exon and some read is spliced
            self.gene.spliced_mol_counts[my_bcidx] += 1
        else:
            self.gene.ambiguous_mol_counts[my_bcidx] += 1



import re
import sys
import gzip
import logging
from typing import *
from collections import defaultdict
import velocyto as vcy
import pysam
import numpy as np

tagpat = re.compile("NH:i:1.+CB:Z:([ACGT]+).+UB:Z:([ACGT]+)")


class ExInCounter:
    """ Main class to do the counting of introns and exons """
    def __init__(self, valid_bcs2idx: Dict[str, int]=None) -> None:
        """ valid_bcs2idx is a dict mapping valid cell barcodes to unique arbitrary numeric indexes used in the output arrays """
        if valid_bcs2idx is None:
            self.valid_bcs2idx: Dict[str, int] = {}
            self.filter_mode = False
            self.counter = 0
        else:
            self.valid_bcs2idx = valid_bcs2idx
            self.filter_mode = True
        self.mask_ivls_by_chromstrand = defaultdict(list)  # type: Dict[str, List]
        # NOTE: by using a default dict and not logging access to keys that do not exist, we might miss bugs!!!
        self.genes = []  # type: List
        self.ivls_by_chromstrand = defaultdict(list)  # type: Dict[str, List]

    @property
    def bclen(self) -> int:
        try:
            return len(next(iter(self.valid_bcs2idx.keys())))
        except StopIteration:
            return None

    def process(self, ivlfile: str, samfile: str) -> None:
        """ Performs all steps in the right order """
        self.read_genes(ivlfile)
        self.mark_up_introns(samfile)
        self.count(samfile)

    @staticmethod
    def parse_cigar_tuple(cigartuples: List[Tuple], pos: int) -> Tuple[List[Tuple[int, int]], bool, int, int]:
        segments = []
        spliced = False
        clip5 = clip3 = 0
        p = pos
        for operation_id, length in cigartuples:
            if operation_id == 0:  # vcy.CIGAR[operation_id] == "BAM_CMATCH"
                segments.append((p, p + length))
                p += length
            elif operation_id == 3:  # A splice || vcy.CIGAR[operation_id] == 'BAM_CREF_SKIP'
                spliced = True
                p += length
            elif operation_id == 2:  # A deletion || cy.CIGAR[operation_id] == 'BAM_CDEL'
                p += length
            elif operation_id == 4:  # bases at 5' or 3' are NOT part of the alignment || vcy.CIGAR[operation_id] == 'BAM_CSOFT_CLIP'
                if p == pos:
                    clip5 = length  # At start of alignment
                else:
                    clip3 = length  # Must be at end of alignment vcy.CIGAR[operation_id] in ["BAM_CINS", "BAM_CHARD_CLIP"]
                p += length
            elif operation_id == 1:  # An insertion BAM_CINS
                pass  # maybe there should be a limit lenght accepted?
            elif operation_id == 5:  # BAM_CHARD_CLIP
                logging.warn("Hard clip was encountered! All mapping are assumed soft clipped")

        return segments, spliced, clip5, clip3

    def iter_unique_alignments(self, samfile: str, yield_line: bool=False) -> Iterable:
        logging.debug(f"Reading {samfile}")
        fin = pysam.AlignmentFile(samfile)  # type: pysam.AlignmentFile
        for i, read in enumerate(fin):
            if read.is_unmapped:
                continue
            try:
                bc = read.get_tag("CB").rstrip("-1")
                umi = read.get_tag("UB")
            except KeyError:
                continue
            if bc not in self.valid_bcs2idx:
                if self.filter_mode:
                    continue
                else:
                    self.valid_bcs2idx[bc] = self.counter
                    self.counter += 1
            strand = '-' if read.is_reverse else '+'
            chrom = fin.get_reference_name(read.rname)  # this is return a string otherwise read.rname for the integer
            if chrom.startswith('chr'):
                # I have to deal with incongruences with the .gft (what is cellranger doing???)
                # NOTE Why this happens?
                if "_" in chrom:
                    chrom = chrom.split("_")[1]
                else:
                    chrom = chrom[3:]
                    if chrom == "M":
                        chrom = "MT"
            pos = read.pos
            segments, spliced, clip5, clip3 = self.parse_cigar_tuple(read.cigartuples, pos)
            if segments == []:
                logging.debug("No segments in read:%s" % read.qname)
            if i % 10000000 == 0:
                logging.debug(f"Read first {i // 1000000} million mapped reads")
            if yield_line:
                yield vcy.Read(bc, umi, chrom, strand, pos, read.cigarstring, segments, clip5, clip3, spliced), read.tostring(fin)
            else:
                yield vcy.Read(bc, umi, chrom, strand, pos, read.cigarstring, segments, clip5, clip3, spliced)
        fin.close()

    def read_repeats(self, gtf_file: str) -> int:
        """ Read masked repeats """
        fin = open(gtf_file)
        for line in fin:
            fields = line.rstrip().split('\t')
            chrom, _, _, start, end, _, strand, _, _ = fields
            # Properly format the chromosome to match how it is parsed in the bam file
            if "_" in chrom:
                chrom = chrom.split("_")[1]
            else:
                chrom = chrom[3:]
                if chrom == "M":
                    chrom = "MT"
            ivlstart = int(start)
            ivlend = int(end)
            chromstrand = chrom + strand
            ivl = vcy.Interval(ivlstart, ivlend, None, 0, "mm10_rmsk", None)
            self.mask_ivls_by_chromstrand[chromstrand].append(ivl)

        # Make sure they are sorted (they should be becouse of the way the files where genrated)
        n = 0
        for chromstrand, ivls in self.mask_ivls_by_chromstrand.items():
            ivls.sort()  # relies on the __lt__ method of vcy.Interval
            n += len(ivls)
        return n

    def read_genes(self, ivlfile: str) -> int:
        """ Read genes and intervals """
        nbcs = len(self.valid_bcs2idx)
        for line in open(ivlfile):
            if line.startswith('#'):
                continue
            fields = line.rstrip().split('\t')
            trname, trid, genename, geneid, chrom, strand = fields[:6]
            if chrom.startswith('chr'):
                chrom = chrom[3:]
            chromstrand = chrom + strand
            ivls = []  # type: List
            gene = vcy.Gene(genename, geneid, chrom, strand, nbcs)
            self.genes.append(gene)
            genestart, geneend = 2**30, 0
            for i in range(6, len(fields)):
                field = fields[i]
                span, ivltype, other_exons, other_introns = field.split(':')
                ivlstart, ivlend = span.split('-')
                ivlstart, ivlend = int(ivlstart), int(ivlend)
                genestart = min(genestart, ivlstart)
                geneend = max(geneend, ivlend)
                is_last_3exon_flag = (i == len(fields) - 1 and strand == "+") or (i == 6 and strand == "-")
                ivl = vcy.Interval(ivlstart, ivlend, gene, len(ivls), ivltype, is_last_3exon_flag)
                self.ivls_by_chromstrand[chromstrand].append(ivl)
                ivls.append(ivl)
            gene.set_range(genestart, geneend)
            gene.set_ivls(ivls)
        n = 0
        for chromstrand, ivls in self.ivls_by_chromstrand.items():
            ivls.sort()  # relies on the __lt__ method of vcy.Interval
            n += len(ivls)
        return n

    def mark_up_introns(self, samfile: str) -> None:
        """ Mark up introns that have reads across junctions to exons ?and count last exon reads? """
        # Read the file
        currchrom = ""
        for r in self.iter_unique_alignments(samfile):
            # Don't consider spliced reads (exonic) in this step
            if r.is_spliced():
                continue

            # When the chromosome mapping of the read changes, change interval index.
            if r.chrom != currchrom:
                logging.debug(f"Marking up chromosome {r.chrom}")
                currchrom = r.chrom
                if currchrom + '+' not in self.ivls_by_chromstrand:
                    logging.warn(f"The .bam file refers to a chromosome '{currchrom}' not present in the annotation (.gtf) file")
                iif = vcy.IntervalsIndex(self.ivls_by_chromstrand[currchrom + '+'])
                iir = vcy.IntervalsIndex(self.ivls_by_chromstrand[currchrom + '-'])
            
            # Consider the correct strand
            ii = iif if r.strand == '+' else iir

            # Look for overlap between the intervals and the read
            matchgenes, matchivls = ii.find_overlapping_ivls(r)
            if len(matchgenes) != 1:  # No match, multiple-matches, chimera, or intron read accidentally spanning another genes' exon zipped between the first genes' exons
                continue

            matchgene = matchgenes.pop()
            matchgene.add_read_stats(r)  # NOTE: this might be removed if the stats output won't be used... but not sure
            
            for matchivl, matchtype in matchivls.items():
                if matchtype == vcy.MATCH_INSIDE:
                    matchivl.add_read_inside()
                if matchtype & vcy.MATCH_OVER5END:  # NOTE: should this be elif? or it makes sense to allow both?
                    matchivl.add_read_spanning5end()
                if matchtype & vcy.MATCH_OVER3END:  # Added recently
                    matchivl.add_read_spanning3end()

        # Validate the intron intervals on the basis of occurrences of Exon/Intron spanning reads
        for gene in self.genes:
            gene.validate_intron_ivls()  # NOTE: this condition might be stricter than we want

        # NOTE: did we properly deal with the reads added to the genes so that self.count() does not get confused?

    def count(self, samfile: str, sam_output: Tuple[Any, Any, Any, Any, Any]=None) -> None:
        """ Do the counting of molecules"""
        molitems = {}  # type: Dict[str, vcy.Molitem]  # keys are the cell_barcode + UMI, values are Molitem objects
        currchrom = ""

        if sam_output is not None:
            f_sure_introns, f_sure_exon, f_maybe_exon, f_others, f_chimeras = sam_output

        # Loop through the aligment of the samfile
        for r in self.iter_unique_alignments(samfile, yield_line=sam_output is not None):
            if sam_output is not None:
                r, line = r

            # When the chromosome mapping of the read changes, change interval index.
            if r.chrom != currchrom:
                currchrom = r.chrom
                logging.debug("Counting on chr %s" % currchrom)
                iif = vcy.IntervalsIndex(self.ivls_by_chromstrand[currchrom + '+'])
                iir = vcy.IntervalsIndex(self.ivls_by_chromstrand[currchrom + '-'])
                iimaskf = vcy.IntervalsIndex(self.mask_ivls_by_chromstrand[currchrom + '+'])
                iimaskr = vcy.IntervalsIndex(self.mask_ivls_by_chromstrand[currchrom + '-'])
                if not (currchrom + '+') in self.mask_ivls_by_chromstrand or not (currchrom + '-') in self.mask_ivls_by_chromstrand:
                    logging.debug(f"{currchrom} is not repeat masked.")

            # Consider the correct strand
            ii = iif if r.strand == '+' else iir
            iim = iimaskf if r.strand == '+' else iimaskr

            # Check if read is fully inside a repeat, in that case skip it
            if iim.has_ivls_enclosing(r):
                continue

            # Look for overlap between the intervals and the read
            matchgenes, matchivls = ii.find_overlapping_ivls(r)
            # logging.debug("matchgenes : %s" % matchgenes)
            if len(matchgenes) != 1:  # No match, multiple-matches, chimera, or intron read accidentally spanning another genes' exon zipped between the first genes' exons
                if sam_output is not None:
                    logging.debug("Chimera inside read or no matching gene")
                    f_chimeras.write(line)
                continue
            
            # Count molecules by dealing with barcodes
            bcumi = r.bc + r.umi
            matchgene = matchgenes.pop()  # NOTE: how does it compare to matchgenes[0]? Is there some gc issue?

            try:
                # Check if the read's gene is the same for the same molecule (bc+umi)
                molitem = molitems[bcumi]
                if molitem.gene != matchgene:
                    # In that case ignore it
                    if sam_output is not None:
                        logging.debug("Chimera between different reads of same molecule")
                        f_chimeras.write(line)
                    continue
            except KeyError:
                # The bc/umi combination was seen for the first time, create a Molecule Item
                molitem = vcy.Molitem(matchgene)
                molitems[bcumi] = molitem

            # Add info about which interval was hit and whether it gives evidence of splicing
            molitem.mark_hit_ivls(matchivls.keys(), r.is_spliced())
            
            if sam_output:
                # Resolve what the read is for the output sam files
                # NOTE: It would be nice to be able to output different sam files by molecule type other than read type
                cum_is_maybe_exon = []
                cum_is_sure_exon = []
                cum_is_sure_intron = []
                for k, v in matchivls.items():
                    cum_is_maybe_exon.append(k.is_maybe_exon)
                    cum_is_sure_exon.append(k.is_sure_exon)
                    cum_is_sure_intron.append(k.is_sure_intron)
                if np.any(cum_is_sure_intron):
                    f_sure_introns.write(line)
                elif np.any(cum_is_sure_exon):
                    f_sure_exon.write(line)
                elif np.any(cum_is_maybe_exon):
                    f_maybe_exon.write(line)

        # After the whole file has been read, do the actual counting
        for bcumi, molitem in molitems.items():
            bc = bcumi[:self.bclen]  # extract the bc part from the bc+umi
            bcidx = self.valid_bcs2idx[bc]
            molitem.count(bcidx)



#!/usr/bin/python
import os
import sys
import shutil
import glob
import click
import re
import time
import logging
import subprocess
logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


@click.command(short_help="Runs the velocity analysis on multiple a Chromium samples in parallel")
@click.argument("parentfolder",
                type=click.Path(exists=True,
                                file_okay=False,
                                dir_okay=True,
                                readable=True,
                                writable=True,
                                resolve_path=True))
@click.argument("ivlfile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.option("--number", "-n",
              help="Number of processes to execute",
              default=5,
              type=click.INT)
@click.option("--wait", "-w",
              help="Delay in seconds between the executions of single run comands",
              default=20,
              type=click.INT)
@click.option("--metadatatable", "-s",
              help="Table containing metadata of the various samples (csv fortmated, [row:samples, col:entry])",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--repmask", "-m",
              help=".gtf file containing intrvals sorted by chromosome, strand, position\n(e.g. generated by running `velocyto extract_repeats mm10_rmsk.gtf`)",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--logfolder", "-l",
              help="Folder where all the log files will be generated",
              default=None,
              type=click.Path(resolve_path=True))
@click.option("--debug", "-d",
              help="debug mode. It will generate .sam files of individual reads (not molecules) that are identified as exons, introns, ambiguous and chimeras",
              default=False,
              is_flag=True)
def multi10x(parentfolder: str, ivlfile: str,
             number: int, wait: int,
             metadatatable: str, repmask: str,
             logfolder: str, debug: bool) -> None:
    """Runs the velocity analysis on multiple a Chromium samples in parallel, spawning several subprocesses
    """
    all_10X_dirs = glob.glob("/data/proj/chromium/10X*")

    logging.debug(f"Attempting to start {number} processes")
    n = 0
    all_10X_dirs.sort()
    for dir_i in all_10X_dirs:
        vc_dir = dir_i + "/velocyto"
        sample_id = dir_i.split("/")[-1]
        name_log_file = os.path.join(logfolder, f'nohup_{sample_id}.out')
        if not os.path.exists(vc_dir):
            try:
                os.mkdir(vc_dir)
            except FileExistsError:
                logging.debug("Wow this the directory was created after we checked! another process is running the sample. Move on.")
                continue
        list_loom = [i for i in os.listdir(vc_dir) if re.match("^10X[_]*?[0-9]{2,3}_[0]*?[1-8]{1,}.loom$", i)]
        if list_loom == [] and not os.path.exists(name_log_file):
            folder = f"/data/proj/chromium/{sample_id}"
            arguments = ["nohup", "velocyto", "run10x", f"{folder}", f"{ivlfile}"]
            if metadatatable:
                arguments += ["--metadatatable", f"{metadatatable}"]
            if repmask:
                arguments += ["--repmask", f"{repmask}"]
            if debug:
                arguments += ["--debug"]
            logging.debug(f"Running the comand {' '.join(arguments)} > {name_log_file} &")
            f_log = open(name_log_file, "w")
            subprocess.Popen(arguments, stdout=f_log, stderr=subprocess.STDOUT)
            n += 1
            if n >= number:
                logging.debug("Done!")
                sys.exit()
            time.sleep(wait)  # seconds
        else:
            logging.debug(f"Skipping {sample_id} appears already running or run.")



import sys
import os
import glob
import re
import gzip
import click
import array
import loompy
import numpy as np
import random
import string
import csv
from collections import defaultdict
import logging
from typing import *
import velocyto as vcy
from ._run import _run

logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


def id_generator(size: int=6, chars: str=string.ascii_uppercase + string.digits) -> str:
    return ''.join(random.choice(chars) for _ in range(size))


@click.command(short_help="Runs the velocity analysis outputing a loom file")
@click.argument("bamfile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.argument("ivlfile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.option("--bcfile", "-b",
              help="""Valid barcodes file, to filter the bam. If --bcfile is not specified all the cell barcodes will be incuded.
              Cell barcodes should be specified in the bcfile as the `CB` tag of each read""",
              default=None,
              show_default=True,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--outputfolder", "-o",
              help="Output folder, does not need to exist",
              default=None,
              type=click.Path(exists=False))
@click.option("--sampleid", "-d",
              help="The sample name that will be used to retrieve informations from metadatatable",
              default=None,
              type=click.Path(exists=False))
@click.option("--metadatatable", "-s",
              help="Table containing metadata of the various samples (csv fortmated, [row:samples, col:entry])",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--repmask", "-m",
              help=".gtf file containing intrvals sorted by chromosome, strand, position\n(e.g. generated by running `velocyto extract_repeats mm10_rmsk.gtf`)",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--debug", "-d",
              help="debug mode. It will generate .sam files of individual reads (not molecules) that are identified as exons, introns, ambiguous and chimeras",
              default=False,
              is_flag=True)
def run(bamfile: str, ivlfile: str,
        bcfile: str, outputfolder: str,
        sampleid: str, metadatatable: str,
        repmask: str, debug: bool,
        additional_ca: dict={}) -> None:
    """Runs the velocity analysis outputing a loom file

    BAMFILE bam file with sorted reads

    IVLFILE text file generated by velocyto extract_intervals
    """
    return _run(bamfile, ivlfile, bcfile, outputfolder,
                sampleid, metadatatable, repmask, debug,
                additional_ca)



import sys
import os
import glob
import re
import gzip
import array
import loompy
import numpy as np
import random
import string
import csv
from collections import defaultdict
import logging
import h5py
from typing import *
import velocyto as vcy

logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


def id_generator(size: int=6, chars: str=string.ascii_uppercase + string.digits) -> str:
    return ''.join(random.choice(chars) for _ in range(size))


def _run(bamfile: str, ivlfile: str,
         bcfile: str, outputfolder: str,
         sampleid: str, metadatatable: str,
         repmask: str, debug: bool,
         additional_ca: dict={}) -> None:
    """Runs the velocity analysis outputing a loom file

    BAMFILE bam file with sorted reads

    IVLFILE text file generated by velocyto extract_intervals
    """
    
    split_sam_flag = debug

    if sampleid is None:
        assert metadatatable is None, "Cannot fetch sample metadata without valid sampleid"
        sampleid = f'{os.path.basename(bamfile).split(".")[0]}_{id_generator(5)}'
        logging.debug(f"No SAMPLEID specified, the sample will be called {sampleid}")

    # Create an output folder inside the cell ranger output folder
    if outputfolder is None:
        outputfolder = os.path.join(os.path.split(bamfile)[0], "velocyto")
        logging.debug(f"No OUTPUTFOLDER specified, find output files inside {outputfolder}")
    if not os.path.exists(outputfolder):
        os.mkdir(outputfolder)

    if bcfile is None:
        logging.debug("Cell barcodes will be determined while reading the .bam file")
    else:
        # Get valid cell barcodes
        valid_bcs_list = [l.strip() for l in open(bcfile).readlines()]
        valid_cellid_list = np.array([f"{sampleid}:{v_bc}" for v_bc in valid_bcs_list])  # with sample id and with -1
        valid_bcs2idx = dict((bc.split('-')[0], n) for n, bc in enumerate(valid_bcs_list))  # without -1
        logging.debug(f"Read {len(valid_bcs_list)} cell barcodes from {bcfile}")
        logging.debug(f"Example of barcode: {valid_bcs_list[0].split('-')[0]} and cell_id: {valid_cellid_list[0]}")
        
    # Get metadata from sample sheet
    if metadatatable:
        try:
            sample_metadata = vcy.MetadataCollection(metadatatable)
            sample = sample_metadata.where("SampleID", sampleid)
            if len(sample) == 0:
                logging.error(f"Sample ID {sampleid} not found in sample sheet")
                # schema = []  # type: List
                sample = {}
            elif len(sample) > 1:
                logging.error(f"Sample ID {sampleid} has multiple lines in sample sheet")
                sys.exit(1)
            else:
                # schema = sample[0].types
                sample = sample[0].dict
            logging.debug(f"Collecting column attributes from {metadatatable}")
        except (NameError, TypeError) as e:
            logging.warn("SAMPLEFILE was not specified. add -s SAMPLEFILE to add metadata.")
            sample = {}
    else:
        sample = {}

    # Initialize Exon-Intron Counter with the valid barcodes
    exincounter = vcy.ExInCounter(valid_bcs2idx)

    # Load the Intervals definition from file
    n = exincounter.read_genes(ivlfile)
    logging.debug(f"Read {n} intervals for {len(exincounter.genes)} genes from {ivlfile}")

    if repmask is not None:
        m = exincounter.read_repeats(repmask)
        logging.debug(f"Read {m} repeat intervals to mask from {repmask}")

    # Go through the sam files a first time to markup introns
    logging.debug("Marking up introns...")
    exincounter.mark_up_introns(bamfile)

    # Do the actual counting
    if split_sam_flag:
        logging.debug("Counting molecules and writing sam outputs...")
        # NOTE: I should write bam file directly using pysam
        f_sure_introns = open(os.path.join(outputfolder, f"{sampleid}_sure_introns.sam"), "w")
        f_sure_exon = open(os.path.join(outputfolder, f"{sampleid}_sure_exon.sam"), "w")
        f_maybe_exon = open(os.path.join(outputfolder, f"{sampleid}_maybe_exon.sam"), "w")
        f_others = open(os.path.join(outputfolder, f"{sampleid}_not_exon_not_intron.sam"), "w")
        f_chimera = open(os.path.join(outputfolder, f"{sampleid}_chimeras.sam"), "w")
        exincounter.count(bamfile, sam_output=(f_sure_introns, f_sure_exon, f_maybe_exon, f_others, f_chimera))
        f_sure_introns.close()
        f_sure_exon.close()
        f_maybe_exon.close()
        f_others.close()
    else:
        logging.debug("Counting molecules...")
        exincounter.count(bamfile)  # NOTE: we would avoid some millions of if statements evalution if we write two function count and count_with output

    if not exincounter.filter_mode:
        valid_bcs2idx = exincounter.valid_bcs2idx  # without -1
        valid_bcs_list = list(zip(*sorted([(v, k) for k, v in valid_bcs2idx.items()])))[1]  # without -1
        valid_cellid_list = np.array([f"{sampleid}:{v_bc}-1" for v_bc in valid_bcs_list])  # with sampleid and with -1
        logging.debug(f"Example of barcode: {valid_bcs_list[0]} and cell_id: {valid_cellid_list[0]}")
    
    ca = {"CellID": np.array(valid_cellid_list)}
    ca.update(additional_ca)

    for key, value in sample.items():
        ca[key] = np.array([value] * len(valid_cellid_list))
    
    # Save 3' junction/exon read counts
    # NOTE: Legacy code this should be added, where is not redunddant to the newer mapstats.hdf5 file
    logging.debug("Save 3' junction/exon read counts")
    olastexon_counts_file = os.path.join(outputfolder, "lastexon_counts.tab")
    ofd = open(olastexon_counts_file, 'w')
    ofd.write("GeneMame\tGeneID\tAnnotatedTrEnd\tDeducedTrEnd\tLastExonLen\tLastJunctionCount\tLastExonCount\tFromEndReadProfile(3'=>5')...\n")
    for g in exincounter.genes:
        lastjunction_count, lastexon_count = g.get_lastexon_counts()
        lastexon_length = g.get_lastexon_length()
        profile = []
        for c in g.read_start_counts_from_locus_end[:lastexon_length]:
            profile.append(c)
        profile_str = "\t".join([str(c) for c in profile])
        ofd.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (g.genename, g.geneid, g.get_tr_end(), g.get_deduced_tr_end(),
                                                        lastexon_length, lastjunction_count, lastexon_count, profile_str))
    ofd.close()

    # Save some stats about exon and introns in a loom file
    logging.debug("Collecting genes structural info statistics")
    
    # Create hdf5 containg the structural stats
    statsfilename = os.path.join(outputfolder, f"{sampleid}_mapstats.hdf5")
    stats_hdf5 = h5py.File(statsfilename, 'w')
    for i, g in enumerate(exincounter.genes):
        # create a group with the Accession Name
        grp = stats_hdf5.create_group(g.geneid)
        type_intervals = np.zeros(len(g.ivls), dtype="|S3")  # not redundand because intron markup is library dependent
        len_intervals = np.zeros(len(g.ivls), dtype="uint32")  # NOTE having this entry to the file is redundant
        valid_intron = np.zeros(len(g.ivls), dtype="bool")
        for j, ivl in enumerate(g.ivls):
            type_intervals[j] = ivl.ivltype
            len_intervals[j] = np.abs(ivl.end - ivl.start)
            valid_intron[j] = ivl.is_sure_valid_intron
        
        grp.create_dataset("reads_per_ivl", data=np.row_stack((g.ivljunction5_read_counts,
                                                               g.ivlinside_read_counts,
                                                               g.ivljunction3_read_counts)))
        grp.create_dataset("ivls_type", data=type_intervals)
        grp.create_dataset("ivls_len", data=len_intervals)
        grp.create_dataset("valid_intron", data=valid_intron)

    stats_hdf5.close()
    logging.debug(f"Mapping statistics have been saved to {statsfilename}")

    # Save to loom file
    outfile = os.path.join(outputfolder, f"{sampleid}.loom")
    logging.debug(f"Generating output file {outfile}")
    
    # row attributes
    atr_table = (("Gene", "genename", str),
                 ("Accession", "geneid", str),
                 ("Chromosome", "chrom", str),
                 ("Strand", "strand", str),
                 ("Start", "start", int),
                 ("End", "end", int))

    logging.debug("Collecting row attributes")
    ra = {}
    for name_col_attr, name_obj_attr, dtyp in atr_table:
        tmp_array = np.zeros((len(exincounter.genes),), dtype=object)  # type: np.ndarray
        for i, g in enumerate(exincounter.genes):
            tmp_array[i] = getattr(g, name_obj_attr)
        ra[name_col_attr] = tmp_array.astype(dtyp)
    
    logging.debug("Generating data table")
    shape_loom = len(exincounter.genes), len(valid_bcs_list)
    spliced = np.zeros(shape_loom, dtype=vcy.LOOM_NUMERIC_DTYPE)
    unspliced = np.zeros(shape_loom, dtype=vcy.LOOM_NUMERIC_DTYPE)
    ambiguous = np.zeros(shape_loom, dtype=vcy.LOOM_NUMERIC_DTYPE)
    for i, g in enumerate(exincounter.genes):
        spliced[i, :] = g.spliced_mol_counts
        unspliced[i, :] = g.unspliced_mol_counts
        ambiguous[i, :] = g.ambiguous_mol_counts
    
    total = spliced + unspliced + ambiguous
    if not np.any(total):
        logging.error("The output file is empty check the input!")

    logging.debug("Writing loom file")
    ds = loompy.create(filename=outfile, matrix=total, row_attrs=ra, col_attrs=ca, dtype="float32")
    ds.set_layer(name="spliced", matrix=spliced, dtype=vcy.LOOM_NUMERIC_DTYPE)
    ds.set_layer(name="unspliced", matrix=unspliced, dtype=vcy.LOOM_NUMERIC_DTYPE)
    ds.set_layer(name="ambiguous", matrix=ambiguous, dtype=vcy.LOOM_NUMERIC_DTYPE)
    ds.close()

    logging.debug("Terminated Succesfully!")
